/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.documents;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.apache.poi.hwpf.HWPFDocument;
import org.apache.poi.hwpf.HWPFOldDocument;
import org.apache.poi.hwpf.OldWordFileFormatException;
import org.apache.poi.hwpf.extractor.Word6Extractor;
import org.apache.poi.hwpf.model.ListData;
import org.apache.poi.hwpf.model.ListLevel;
import org.apache.poi.hwpf.model.ListTables;
import org.apache.poi.hwpf.model.PicturesTable;
import org.apache.poi.hwpf.model.StyleDescription;
import org.apache.poi.hwpf.usermodel.CharacterRun;
import org.apache.poi.hwpf.usermodel.Paragraph;
import org.apache.poi.hwpf.usermodel.Picture;
import org.apache.poi.hwpf.usermodel.Range;
import org.apache.poi.hwpf.usermodel.Table;
import org.apache.poi.hwpf.usermodel.TableCell;
import org.apache.poi.hwpf.usermodel.TableRow;
import org.apache.poi.poifs.filesystem.DirectoryEntry;
import org.apache.poi.poifs.filesystem.Entry;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.tika.exception.TikaException;
import org.apache.tika.io.TikaInputStream;
import org.apache.tika.parser.ParseContext;
import org.apache.tika.parser.microsoft.WordExtractor;
import org.apache.tika.parser.microsoft.OfficeParser.POIFSDocumentType;
import org.apache.tika.sax.XHTMLContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

public class WordDocExtractor extends WordExtractor {

    private POIFSFileSystem fs;
    public static final float TWIPS_PER_INCH = 1440.0f;
    public static final int TWIPS_PER_PT = 20;
    private boolean isListOpened = false;
    private Stack<String> listTag = new Stack<String>();
    private int lastParagraphLevel = 0;

    public WordDocExtractor(ParseContext context) {
	super(context);
    }

    @Override
    protected void parse(POIFSFileSystem filesystem, XHTMLContentHandler xhtml)
	    throws IOException, SAXException, TikaException {
	fs = filesystem;
	HWPFDocument document;
	try {
	    document = new HWPFDocument(filesystem);
	} catch (OldWordFileFormatException e) {
	    parseWord6(filesystem, xhtml);
	    return;
	}
	org.apache.poi.hwpf.extractor.WordExtractor wordExtractor = new org.apache.poi.hwpf.extractor.WordExtractor(
		document);

	addTextIfAny(xhtml, "header", wordExtractor.getHeaderText());
	PicturesTable pictureTable = document.getPicturesTable();
	PicturesSource pictures = new PicturesSource(document);

	Range r = document.getRange();
	for (int i = 0; i < r.numParagraphs(); i++) {
	    Paragraph p = r.getParagraph(i);
	    i += handleParagraph(p, 0, r, document, pictures, pictureTable,
		    xhtml);
	}

	for (String paragraph : wordExtractor.getFootnoteText()) {
	    xhtml.element("p", paragraph);
	}

	for (String paragraph : wordExtractor.getCommentsText()) {
	    xhtml.element("p", paragraph);
	}

	for (String paragraph : wordExtractor.getEndnoteText()) {
	    xhtml.element("p", paragraph);
	}

	addTextIfAny(xhtml, "footer", wordExtractor.getFooterText());

	for (Picture p = pictures.nextUnclaimed(); p != null;) {
	    handlePictureCharacterRun(null, p, pictures, xhtml);
	    p = pictures.nextUnclaimed();
	}
    }

    private int handleParagraph(Paragraph p, int parentTableLevel, Range r,
	    HWPFDocument document, PicturesSource pictures,
	    PicturesTable pictureTable, XHTMLContentHandler xhtml)
	    throws SAXException, IOException, TikaException {
	if (p.getIlfo() != 0) {
	    int level = p.getIlvl();
	    if (!isListOpened || (isListOpened && level > lastParagraphLevel)) {
		String tag;
		if(document.getListTables().getListData(document.getListTables().getOverride(p.getIlfo()).getLsid()).getLevels()[level].getNumberFormat()==0)
		    tag = "ol";
		else
		    tag = "ul";
		xhtml.startElement(tag);
		listTag.push(tag);
		isListOpened = true;
	    }
	    if (isListOpened && level < lastParagraphLevel)
		for(int i=0; i<lastParagraphLevel-level; i++){
		    String tag = listTag.pop();
		    xhtml.endElement(tag);
		    if(listTag.empty()){
			isListOpened = false;
			break;
		    }
		}
	    lastParagraphLevel = level;
	    xhtml.startElement("li");
	    for (int j = 0; j < p.numCharacterRuns(); j++) {
		CharacterRun cr = p.getCharacterRun(j);
		if (cr.text().equals("\u0013")) {
		    j += handleSpecialCharacterRuns(p, j, false, pictures,
			    xhtml);
		} else if (cr.text().startsWith("\u0008")) {
		    for (int pn = 0; pn < cr.text().length(); pn++) {
			Picture picture = pictures.nextUnclaimed();
			handlePictureCharacterRun(cr, picture, pictures, xhtml);
		    }
		} else if (pictureTable.hasPicture(cr)) {
		    Picture picture = pictures.getFor(cr);
		    handlePictureCharacterRun(cr, picture, pictures, xhtml);
		} else {
		    handleCharacterRun(cr, false, xhtml);
		}
	    }
	    xhtml.endElement("li");
	    return 0;
	} else if (isListOpened){
	    String tag = listTag.pop();
	    xhtml.endElement(tag);
	    isListOpened = false;
	}

	if (p.isInTable() && p.getTableLevel() > parentTableLevel
		&& parentTableLevel == 0) {
	    Table t = r.getTable(p);
	    xhtml.startElement("table");
	    xhtml.startElement("tbody");
	    for (int rn = 0; rn < t.numRows(); rn++) {
		TableRow row = t.getRow(rn);
		xhtml.startElement("tr");
		for (int cn = 0; cn < row.numCells(); cn++) {
		    TableCell cell = row.getCell(cn);
		    xhtml.startElement("td");

		    for (int pn = 0; pn < cell.numParagraphs(); pn++) {
			Paragraph cellP = cell.getParagraph(pn);
			handleParagraph(cellP, p.getTableLevel(), cell,
				document, pictures, pictureTable, xhtml);
		    }
		    xhtml.endElement("td");
		}
		xhtml.endElement("tr");
	    }
	    xhtml.endElement("tbody");
	    xhtml.endElement("table");
	    return (t.numParagraphs() - 1);
	}

	/*if (p.getIlfo() != 0) { // lists handling
	    ListTables lists = document.getListTables();
	    ListData data = lists.getListData(lists.getOverride(p.getIlfo())
		    .getLsid());
	    ListLevel[] levels = data.getLevels();
	    ListLevel level = levels[p.getIlvl()];
	    System.out.println("Lsid: "
		    + lists.getOverride(p.getIlfo()).getLsid());
	    System.out.println("Text: " + p.text());
	    System.out.println("Level: " + p.getIlvl());
	    System.out.println("NumberText: " + level.getNumberText());
	    System.out.println("NumberFormat: " + level.getNumberFormat());
	}*/

	StyleDescription style = document.getStyleSheet().getStyleDescription(
		p.getStyleIndex());
	TagAndStyle tas = buildParagraphTagAndStyle(style.getName(),
		(parentTableLevel > 0));

	if (tas.getStyleClass() != null) {
	    xhtml.startElement(tas.getTag(), "class", tas.getStyleClass());
	} else {
	    xhtml.startElement(tas.getTag());
	}

	for (int j = 0; j < p.numCharacterRuns(); j++) {
	    CharacterRun cr = p.getCharacterRun(j);
	    if (cr.text().equals("\u0013")) {
		j += handleSpecialCharacterRuns(p, j, tas.isHeading(),
			pictures, xhtml);
	    } else if (cr.text().startsWith("\u0008")) {
		for (int pn = 0; pn < cr.text().length(); pn++) {
		    Picture picture = pictures.nextUnclaimed();
		    handlePictureCharacterRun(cr, picture, pictures, xhtml);
		}
	    } else if (pictureTable.hasPicture(cr)) {
		Picture picture = pictures.getFor(cr);
		handlePictureCharacterRun(cr, picture, pictures, xhtml);
	    } else {
		handleCharacterRun(cr, tas.isHeading(), xhtml);
	    }
	}

	xhtml.endElement(tas.getTag());

	return 0;
    }

    private void handleCharacterRun(CharacterRun cr, boolean skipStyling,
	    XHTMLContentHandler xhtml) throws SAXException {
	if (cr.text().equals("\r"))
	    return;

	List<String> tags = new ArrayList<String>();
	if (!skipStyling) {
	    if (cr.isBold())
		tags.add("b");
	    if (cr.isItalic())
		tags.add("i");
	    if (cr.isStrikeThrough())
		tags.add("s");
	    switch (cr.getSubSuperScriptIndex()) {
	    case 0:
		break;
	    case 1:
		tags.add("sup");
		break;
	    case 2:
		tags.add("sub");
		break;
	    }
	    for (String tag : tags) {
		xhtml.startElement(tag);
	    }
	}
	String text = cr.text();
	text = text.replace('\r', '\n');
	if (text.endsWith("\u0007")) {
	    text = text.substring(0, text.length() - 1);
	}

	xhtml.characters(text);

	for (int tn = tags.size() - 1; tn >= 0; tn--) {
	    xhtml.endElement(tags.get(tn));
	}
    }

    private int handleSpecialCharacterRuns(Paragraph p, int index,
	    boolean skipStyling, PicturesSource pictures,
	    XHTMLContentHandler xhtml) throws SAXException, TikaException,
	    IOException {
	List<CharacterRun> controls = new ArrayList<CharacterRun>();
	List<CharacterRun> texts = new ArrayList<CharacterRun>();
	boolean has14 = false;
	boolean parsingSuccess = false;

	int i;
	for (i = index + 1; i < p.numCharacterRuns(); i++) {
	    CharacterRun cr = p.getCharacterRun(i);
	    if (cr.text().equals("\u0013")) {
		int increment = handleSpecialCharacterRuns(p, i + 1,
			skipStyling, pictures, xhtml);
		i += increment;
	    } else if (cr.text().equals("\u0014")) {
		has14 = true;
		if (cr.isOle2()) {
		    parsingSuccess = parseEmbeddedWord(xhtml,
			    "" + cr.getPicOffset());
		}
	    } else if (cr.text().equals("\u0015")) {
		if (!has14) {
		    texts = controls;
		    controls = new ArrayList<CharacterRun>();
		}
		break;
	    } else {
		if (has14) {
		    texts.add(cr);
		} else {
		    controls.add(cr);
		}
	    }
	}

	if (controls.size() > 0) {
	    String text = controls.get(0).text();
	    for (int j = 1; j < controls.size(); j++) {
		text += controls.get(j).text();
	    }

	    if (text.startsWith("HYPERLINK") && text.indexOf('"') > -1) {
		String url = text.substring(text.indexOf('"') + 1,
			text.lastIndexOf('"'));
		xhtml.startElement("a", "href", url);
		for (CharacterRun cr : texts) {
		    handleCharacterRun(cr, skipStyling, xhtml);
		}
		xhtml.endElement("a");
	    } else if (!parsingSuccess) {
		for (CharacterRun cr : texts) {
		    if (pictures.hasPicture(cr)) {
			Picture picture = pictures.getFor(cr);
			handlePictureCharacterRun(cr, picture, pictures, xhtml);
		    } else {
			handleCharacterRun(cr, skipStyling, xhtml);
		    }
		}
	    }
	} else {
	    for (CharacterRun cr : texts) {
		handleCharacterRun(cr, skipStyling, xhtml);
	    }
	}

	return i - index;
    }

    private boolean parseEmbeddedWord(XHTMLContentHandler xhtml, String offset)
	    throws TikaException, SAXException, IOException {
	try {
	    DirectoryEntry objectPool = (DirectoryEntry) fs.getRoot().getEntry(
		    "ObjectPool");
	    for (Entry entry : objectPool) {
		if (entry.getName().startsWith("_")
			&& entry.getName().regionMatches(1, offset, 0,
				offset.length())
			&& entry instanceof DirectoryEntry) {
		    if (canHandle((DirectoryEntry) entry)) {
			handleEmbededOfficeDoc((DirectoryEntry) entry, xhtml);
			return true;
		    } else
			return false;
		}
	    }
	    return false;
	} catch (FileNotFoundException e) {
	    return false;
	}
    }

    private boolean canHandle(DirectoryEntry dir) throws IOException {
	POIFSDocumentType type = POIFSDocumentType.detectType(dir);
	if (type != POIFSDocumentType.WORDDOCUMENT)
	    return false;
	return true;
    }

    private void handlePictureCharacterRun(CharacterRun cr, Picture picture,
	    PicturesSource pictures, XHTMLContentHandler xhtml)
	    throws SAXException, IOException, TikaException {
	if (picture == null) {
	    return;
	}

	String extension = picture.suggestFileExtension();
	int pictureNumber = pictures.pictureNumber(picture);

	String filename = "image" + pictureNumber
		+ (extension.length() > 0 ? "." + extension : "");

	String mimeType = picture.getMimeType();

	final int aspectRatioX = picture.getHorizontalScalingFactor();
	final int aspectRatioY = picture.getVerticalScalingFactor();

	final float imageWidth;
	final float imageHeight;

	final float cropTop;
	final float cropBottom;
	final float cropLeft;
	final float cropRight;

	if (aspectRatioX > 0) {
	    imageWidth = picture.getDxaGoal() * aspectRatioX / 1000
		    / TWIPS_PER_PT;
	    cropRight = picture.getDxaCropRight() * aspectRatioX / 1000
		    / TWIPS_PER_PT;
	    cropLeft = picture.getDxaCropLeft() * aspectRatioX / 1000
		    / TWIPS_PER_PT;
	} else {
	    imageWidth = picture.getDxaGoal() / TWIPS_PER_PT;
	    cropRight = picture.getDxaCropRight() / TWIPS_PER_PT;
	    cropLeft = picture.getDxaCropLeft() / TWIPS_PER_PT;
	}

	if (aspectRatioY > 0) {
	    imageHeight = picture.getDyaGoal() * aspectRatioY / 1000
		    / TWIPS_PER_PT;
	    cropTop = picture.getDyaCropTop() * aspectRatioY / 1000
		    / TWIPS_PER_PT;
	    cropBottom = picture.getDyaCropBottom() * aspectRatioY / 1000
		    / TWIPS_PER_PT;
	} else {
	    imageHeight = picture.getDyaGoal() / TWIPS_PER_PT;
	    cropTop = picture.getDyaCropTop() / TWIPS_PER_PT;
	    cropBottom = picture.getDyaCropBottom() / TWIPS_PER_PT;
	}
	if (cr != null && mimeType.contains("wmf")) {
	    if (cropTop != 0 || cropRight != 0 || cropBottom != 0
		    || cropLeft != 0) {
		float visibleWidth = Math.max(0, imageWidth - cropLeft
			- cropRight);
		float visibleHeight = Math.max(0, imageHeight - cropTop
			- cropBottom);
		filename = "image" + pictureNumber + ".svg";
		AttributesImpl attributes = new AttributesImpl();
		attributes.addAttribute("", "data", "data", "", "embedded:"
			+ filename);
		attributes.addAttribute("", "height", "height", "", ""
			+ visibleHeight);
		attributes.addAttribute("", "width", "width", "", ""
			+ visibleWidth);
		attributes
			.addAttribute("", "type", "type", "", "image/svg+xml"); // it
										// would
										// be
										// converted
										// to
										// svg
										// later
		xhtml.startElement("object", attributes);
	    } else {
		filename = "image" + pictureNumber + ".svg";
		AttributesImpl attributes = new AttributesImpl();
		attributes.addAttribute("", "data", "data", "", "embedded:"
			+ filename);
		attributes.addAttribute("", "height", "height", "", ""
			+ imageHeight);
		attributes.addAttribute("", "width", "width", "", ""
			+ imageWidth);
		attributes
			.addAttribute("", "type", "type", "", "image/svg+xml"); // it
										// would
										// be
										// converted
										// to
										// svg
										// later
		xhtml.startElement("object", attributes);
	    }
	} else
	    xhtml.startElement("img", "src", "embedded:" + filename);
	xhtml.endElement("img");

	if (!pictures.hasOutput(picture)) {
	    TikaInputStream stream = TikaInputStream.get(picture.getContent());
	    handleEmbeddedResource(stream, filename, mimeType, xhtml, false);
	    pictures.recordOutput(picture);
	}
    }

    private void addTextIfAny(XHTMLContentHandler xhtml, String section,
	    String text) throws SAXException {
	if (text != null && text.length() > 0) {
	    xhtml.startElement("div", "class", section);
	    xhtml.element("p", text);
	    xhtml.endElement("div");
	}
    }

    protected void parseWord6(POIFSFileSystem filesystem,
	    XHTMLContentHandler xhtml) throws IOException, SAXException,
	    TikaException {
	HWPFOldDocument doc = new HWPFOldDocument(filesystem);
	Word6Extractor extractor = new Word6Extractor(doc);

	for (String p : extractor.getParagraphText()) {
	    xhtml.element("p", p);
	}
    }

    public static TagAndStyle buildParagraphTagAndStyle(String styleName,
	    boolean isTable) {
	String tag = "p";
	String styleClass = null;

	if (styleName.equals("Default") || styleName.equals("Normal")) {
	} else if (styleName.equals("Table Contents") && isTable) {
	} else if (styleName.equals("heading") || styleName.equals("Heading")) {
	    tag = "h1";
	} else if (styleName.startsWith("heading")
		|| styleName.startsWith("Heading")) {
	    int num = 1;
	    try {
		num = Integer
			.parseInt(styleName.substring(styleName.length() - 1));
	    } catch (NumberFormatException e) {
	    }
	    tag = "h" + num;
	} else if (styleName.equals("Title")) {
	    tag = "h1";
	    styleClass = "title";
	} else if (styleName.equals("Subtitle")) {
	    tag = "h2";
	    styleClass = "subtitle";
	} else if (styleName.equals("HTML Preformatted")) {
	    tag = "pre";
	} else {
	    styleClass = styleName.replace(' ', '_');
	    styleClass = styleClass.substring(0, 1).toLowerCase()
		    + styleClass.substring(1);
	}

	return new TagAndStyle(tag, styleClass);
    }

    private static class PicturesSource {
	private PicturesTable picturesTable;
	private Set<Picture> output = new HashSet<Picture>();
	private Map<Integer, Picture> lookup;
	private List<Picture> nonU1based;
	private List<Picture> all;
	private int pn = 0;

	private PicturesSource(HWPFDocument doc) {
	    picturesTable = doc.getPicturesTable();
	    all = picturesTable.getAllPictures();

	    lookup = new HashMap<Integer, Picture>();
	    for (Picture p : all) {
		String name = p.suggestFullFileName();
		if (name.indexOf('.') > -1)
		    name = name.substring(0, name.indexOf('.'));
		int offset = Integer.parseInt(name, 16);
		lookup.put(offset, p);
	    }
	    nonU1based = new ArrayList<Picture>();
	    nonU1based.addAll(all);
	    Range r = doc.getRange();
	    for (int i = 0; i < r.numCharacterRuns(); i++) {
		CharacterRun cr = r.getCharacterRun(i);
		if (picturesTable.hasPicture(cr)) {
		    Picture p = getFor(cr);
		    int at = nonU1based.indexOf(p);
		    nonU1based.set(at, null);
		}
	    }
	}

	private boolean hasPicture(CharacterRun cr) {
	    return picturesTable.hasPicture(cr);
	}

	private void recordOutput(Picture picture) {
	    output.add(picture);
	}

	private boolean hasOutput(Picture picture) {
	    return output.contains(picture);
	}

	private int pictureNumber(Picture picture) {
	    return all.indexOf(picture) + 1;
	}

	private Picture getFor(CharacterRun cr) {
	    return lookup.get(cr.getPicOffset());
	}

	private Picture nextUnclaimed() {
	    Picture p = null;
	    while (pn < nonU1based.size()) {
		p = nonU1based.get(pn);
		pn++;
		if (p != null)
		    return p;
	    }
	    return null;
	}
    }
}
