/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.IDocumentProcessor;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.reports.ITemplatePathSource;
import com.unitesk.requality.eclipse.tools.IUIStateChangeListener;
import com.unitesk.requality.eclipse.tools.TreesTracker;
import com.unitesk.requality.report.IReportDataSource;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;

/**
 * The activator class controls the plug-in life cycle
 */
public class Activator extends AbstractUIPlugin {

    // The plug-in ID
    public static final String PLUGIN_ID = "com.unitesk.requality";

    public static TreesTracker tracker = TreesTracker.getTracker();

    public static final String EXTENSION_NAME_DOC_PROCESSOR = "com.unitesk.requality.document.processor";
    public static final String EXTENSION_NAME_REPORT = "com.unitesk.requality.report";
    public static final String EXTENSION_NAME_NODE_TYPES = "com.unitesk.requality.nodetype";

    // UI state change listener
    private List<IUIStateChangeListener> listeners = new ArrayList<IUIStateChangeListener>();

    private Map<String, TreeNode> activeNodes = new HashMap<String, TreeNode>();

    // The shared instance
    private static Activator plugin;

    IResourceChangeListener resourceChangeListener = new IResourceChangeListener() {
        @Override
        public void resourceChanged(IResourceChangeEvent event) {
            if (event.getType() == IResourceChangeEvent.PRE_CLOSE) {
                tracker.stopTracking((IProject) event.getResource(), false);
                return;
            }
            if (event.getType() == IResourceChangeEvent.PRE_DELETE) {
                tracker.stopTracking((IProject) event.getResource(), true);
                return;
            }
            if (event.getType() == IResourceChangeEvent.POST_CHANGE) {
                // Track new
                IProject newProjects[] = getNewProjects();
                for (IProject proj : newProjects) {
                    tracker.trackProject(proj);
                }
            }
        }
    };

    /**
     * The constructor
     */
    public Activator() {
        plugin = this;
    }

    private SortedSet<DocumentProcessorDescriptor> processors = new TreeSet<DocumentProcessorDescriptor>();
    private Map<String, Template> reportTemplates = new HashMap<String, Template>();

    public SortedSet<DocumentProcessorDescriptor> getProcessors() {
        loadDocumentProcessorDescriptors();
        return Collections.unmodifiableSortedSet(processors);
    }

    public Map<String, Template> getReportTemplates() {
        loadReportTemplatesDescriptors();
        return Collections.unmodifiableMap(reportTemplates);
    }

    private void loadReportTemplatesDescriptors() {
        List<String> tplPaths = new ArrayList<String>();
        try {
            tplPaths.add(Activator.getPath("templates"));
        } catch (IOException e) {
            e.printStackTrace();
        }
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_REPORT);
        for (IConfigurationElement conf : configurationElements) {
            try {
                if (!conf.getName().equals("templatePath")) {
                    continue;
                }
                ITemplatePathSource tplPath = (ITemplatePathSource) conf
                        .createExecutableExtension("provider");
                String path = tplPath.getPath();
                if (path != null && new File(path).isDirectory()) {
                    tplPaths.add(path);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        for (String path : tplPaths) {
            File tplPath = new File(path);
            for (String tplFolder : tplPath.list()) {

                if (new File(tplPath + File.separator + tplFolder
                        + File.separator + "template.xml").exists()) {
                    Template template = new Template(tplPath + File.separator
                            + tplFolder);
                    reportTemplates.put(template.getId(), template);
                }
            }
        }
        for (IConfigurationElement conf : configurationElements) {
            if (!conf.getName().equals("dataProvider")) {
                continue;
            }
            String tId = conf.getAttribute("templateId");
            String ftlName = conf.getAttribute("ftlName");
            if (reportTemplates.containsKey(tId)) {
                Template tpl = reportTemplates.get(tId);
                for (TemplateFile t : tpl.getTemplates()) {
                    if (t.getTemplatePath().equals(ftlName)) {
                        IReportDataSource dataObject;
                        try {
                            dataObject = (IReportDataSource) conf
                                    .createExecutableExtension("dataSource");
                            t.setDataObject(dataObject);
                        } catch (CoreException e) {
                            e.printStackTrace();
                        }
                    }
                }
            }
        }
    }

    public void loadDocumentProcessorDescriptors() {
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_DOC_PROCESSOR);
        for (IConfigurationElement conf : configurationElements) {
            try {
                IDocumentProcessor docProc = (IDocumentProcessor) conf
                        .createExecutableExtension("class");
                processors
                        .add(new DocumentProcessorDescriptor(conf
                                .getAttribute("id"), conf.getAttribute("name"),
                                docProc));

            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext
     * )
     */
    public void start(BundleContext context) throws Exception {
        super.start(context);
        ResourcesPlugin.getWorkspace().addResourceChangeListener(
                resourceChangeListener);
        initTracker();
    }

    public String getJSToolsPath() {
        String path = null;
        try {
            path = FileLocator.toFileURL(this.getBundle().getEntry("jstools"))
                    .getFile();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return path;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext
     * )
     */
    public void stop(BundleContext context) throws Exception {
        plugin = null;
        super.stop(context);
    }

    /**
     * Returns the shared instance
     * 
     * @return the shared instance
     */
    public static Activator getDefault() {
        return plugin;
    }

    @Override
    protected ImageRegistry createImageRegistry() {
        ImageRegistry imgreg = super.createImageRegistry();

        Map<String, NodeTypeDescriptor> types = RequalityPlugin
                .getAdditionalTypes();
        for (String n : types.keySet()) {
            NodeTypeDescriptor desc = RequalityPlugin.getAdditionalTypes().get(
                    n);
            if (desc.getIcon() != null) {
                imgreg.put(
                        "nodetype." + n,
                        getImageDescriptor(desc.getPluginId(), RequalityPlugin
                                .getAdditionalTypes().get(n).getIcon()));
            } else {
                imgreg.put(
                        "nodetype." + n,
                        getImageDescriptor(desc.getPluginId(),
                                "icons/unitesk.png"));
            }
        }
        imgreg.put("nodetype.NodeType",
                getImageDescriptor("icons/nodetypes/NodeType.png"));
        imgreg.put("nodetype.default", getImageDescriptor("icons/unitesk.png"));
        return imgreg;
    }

    private static IProject[] getNewProjects() {
        TreesTracker tracker = TreesTracker.getTracker();
        Set<IProject> openProjects = new HashSet<IProject>();

        for (TreeDB tree : tracker.getTrees())
            if (tracker.getProject(tree).isOpen()) {
                openProjects.add(tracker.getProject(tree));
            }
        Set<IProject> curProjects = new HashSet<IProject>();
        for (IProject proj : ResourcesPlugin.getWorkspace().getRoot()
                .getProjects()) {
            if (isInterestingProject(proj)) {
                curProjects.add(proj);
            }
        }

        Set<IProject> result = new HashSet<IProject>(curProjects);
        result.removeAll(openProjects);
        return result.toArray(new IProject[0]);
    }

    private static boolean isInterestingProject(IProject proj) {
        try {
            if (!proj.isOpen()) {
                return false;
            }
            if (!proj.isNatureEnabled(ReqCatalogue.ID)) {
                return false;
            }
        } catch (CoreException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    private static void initTracker() throws CoreException {
        LinkedList<IProject> reqProjects = new LinkedList<IProject>();
        for (IProject changedProject : ResourcesPlugin.getWorkspace().getRoot()
                .getProjects()) {
            if (isInterestingProject(changedProject)) {
                reqProjects.add(changedProject);
            }
        }
        for (IProject changedProject : reqProjects) {
            if (!TreesTracker.getTracker().isTracking(changedProject)) {
                TreesTracker.getTracker().trackProject(changedProject);
            }
        }
    }

    public static String getPath(String string) throws IOException {
        return FileLocator.toFileURL(
                Activator.getDefault().getBundle().getResource(string))
                .getFile();
    }

    /**
     * Returns an image descriptor for the image file at the given plug-in
     * relative path
     * 
     * @param path
     *            the path
     * @return the image descriptor
     */
    public static ImageDescriptor getImageDescriptor(String pluginId,
            String path) {
        return imageDescriptorFromPlugin(pluginId, path);
    }

    public static ImageDescriptor getImageDescriptor(String path) {
        return imageDescriptorFromPlugin(PLUGIN_ID, path);
    }

    public synchronized boolean addChangeListener(
            IUIStateChangeListener listener) {
        return listeners.add(listener);
    }

    public synchronized boolean removeChangeListener(
            IUIStateChangeListener listener) {
        return listeners.remove(listener);
    }

    public synchronized void setActiveNode(TreeNode node, String type) {
        activeNodes.put(type, node);
        for (IUIStateChangeListener listener : listeners) {
            listener.newActiveNode(node);
        }
    }

    public synchronized TreeNode getActiveNode(String string) {
        return activeNodes.get(string);
    }

}
