package com.unitesk.requality.merge;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.unitesk.requality.merge.Chunk.InvalidLocationChunkException;
import com.unitesk.requality.merge.Location.TraceFailedException;

public class MergeProcessor {

    private static final MergeProcessor instance = new MergeProcessor();

    private MergeProcessor() {
    }

    public static MergeProcessor getInstance() {
	return instance;
    }

    public List<String> merge(Document v1, Document v2) {
	Tree t1 = new Tree(v1);
	Tree t2 = new Tree(v2);
	List<Location> oldLocations = lookupLocations(t1);
	List<Location> newLocations = traceLocations(oldLocations, t2);
	Map<Node, List<Chunk>> locationMapping = mapLocations(newLocations);
	Map<Node, Fragment> fragmentMapping = makeFragments(locationMapping);
	Iterator<Node> i = fragmentMapping.keySet().iterator();
	while (i.hasNext()) {
	    Node anchorNode = i.next();
	    mergeFragment(fragmentMapping.get(anchorNode), anchorNode);
	    t2.updateTree();
	}
	List<String> mergedIds = new ArrayList<String>();
	for (Location loc : newLocations)
	    mergedIds.add(loc.getLocationId());
	return mergedIds;
    }

    private Map<Node, Fragment> makeFragments(
	    Map<Node, List<Chunk>> locationMapping) {
	Map<Node, Fragment> mapping = new HashMap<Node, Fragment>();
	Iterator<Node> i = locationMapping.keySet().iterator();
	while (i.hasNext()) {
	    Node anchorNode = i.next();
	    List<Chunk> locations = locationMapping.get(anchorNode);
	    List<Node> relatedNodes = new ArrayList<Node>();
	    for (Chunk loc : locations) {
		List<Node> locRelatedNodes = new ArrayList<Node>();
		Node spanNode = loc.getSpanNode();
		Node leftSibling = spanNode.getPreviousSibling();
		Node rightSibling = spanNode.getNextSibling();
		if (leftSibling != null && leftSibling.isText())
		    locRelatedNodes.add(leftSibling);
		locRelatedNodes.add(spanNode);
		if (rightSibling != null && rightSibling.isText())
		    locRelatedNodes.add(rightSibling);
		for (Node x : locRelatedNodes)
		    if (!relatedNodes.contains(x))
			relatedNodes.add(x);
	    }
	    Fragment f = new Fragment();
	    for (Node x : relatedNodes)
		f.attachNode(x, true);
	    mapping.put(anchorNode, f);
	}
	return mapping;
    }

    private void mergeFragment(Fragment f, Node anchorNode) {
	org.w3c.dom.Node anchorDomNode = anchorNode.getDomNode();
	org.w3c.dom.Node fRoot = f.getRoot().getDomNode();
	fRoot = anchorDomNode.getOwnerDocument().importNode(fRoot, true);
	anchorDomNode.getParentNode().replaceChild(fRoot, anchorDomNode);
    }

    private Map<Node, List<Chunk>> mapLocations(List<Location> locations) {
	Map<Node, List<Chunk>> mapping = new HashMap<Node, List<Chunk>>();
	Map<Node, List<Chunk>> ctxMapping = new HashMap<Node, List<Chunk>>();

	for (Location loc : locations) {
	    Node cp = loc.getCommonParent();
	    if (!ctxMapping.containsKey(cp))
		ctxMapping.put(cp, loc.getChunks());
	    else {
		List<Chunk> chunks = ctxMapping.get(cp);
		chunks.addAll(loc.getChunks());
		ctxMapping.put(cp, chunks);
	    }
	}
	Iterator<Node> i = ctxMapping.keySet().iterator();
	while (i.hasNext()) {
	    Node cp = i.next();
	    List<Chunk> chunks = ctxMapping.get(cp);
	    if (chunks.size() == 1) {
		Node anchorNode = chunks.get(0).getAnchorNode();
		mapping.put(anchorNode, chunks);
	    } else {
		for (Chunk chunk : chunks) {
		    Node anchorNode = chunk.getAnchorNode();
		    if (mapping.containsKey(anchorNode))
			mapping.get(anchorNode).add(chunk);
		    else {
			List<Chunk> list = new ArrayList<Chunk>();
			list.add(chunk);
			mapping.put(anchorNode, list);
		    }
		}
	    }
	}
	return mapping;
    }

    private List<Location> traceLocations(List<Location> locations, Tree t) {
	List<Location> newLocations = new ArrayList<Location>();
	for (Location loc : locations) {
	    try {
		Location newLoc = loc.traceLocation(t);
		newLocations.add(newLoc);
	    } catch (TraceFailedException e) {
		switch (loc.getStatus()) {
		case MISSING:
		    System.out.println("Location " + loc.getLocationId()
			    + " is missing completely");
		    break;
		case MOVED:
		    System.out.println("Location " + loc.getLocationId()
			    + " is moved");
		    System.out.println("Was under " + e.getExpectedValue());
		    System.out.println("Now under " + e.getActualValue());
		    break;
		case CHANGED:
		    System.out.println("Location " + loc.getLocationId()
			    + " is changed");
		    System.out.println("Expected: " + e.getExpectedValue());
		    System.out.println("Actual: " + e.getActualValue());
		    break;
		default:
		    System.out.println("Ok, that shouldn't happen. WTF?");
		    break;
		}
	    }
	}
	return newLocations;
    }

    public String[] merge(Reader v1, Reader v2, Writer output) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory
		    .newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc1 = builder.parse(new InputSource(v1));
	    Document doc2 = builder.parse(new InputSource(v2));
	    List<String> result;
	    try {
		result = merge(doc1, doc2);
	    } catch (NullPointerException e) {
		e.printStackTrace();
		throw e;
	    } finally {
		Utils.writeXmlFile(doc2, output);
	    }
	    return result.toArray(new String[0]);
	} catch (ParserConfigurationException e) {
	    e.printStackTrace();
	} catch (SAXException e) {
	    e.printStackTrace();
	} catch (IOException e) {
	    e.printStackTrace();
	}
	return new String[0];
    }

    private List<Location> lookupLocations(Tree t) {
	List<Location> locations = new ArrayList<Location>();
	List<String> idsFound = new ArrayList<String>();
	for (Node x : t.getTraversalSequence()) {
	    if (Chunk.isLocation(x))
		try {
		    Chunk chunk = new Chunk(x);
		    if (idsFound.contains(chunk.getLocationId())) {
			for (Location loc : locations)
			    if (loc.getLocationId().equals(
				    chunk.getLocationId()))
				loc.addChunk(chunk);
		    } else {
			Location loc = new Location(chunk);
			locations.add(loc);
			idsFound.add(loc.getLocationId());
		    }
		} catch (InvalidLocationChunkException e) {
		    e.printStackTrace();
		}
	}
	return locations;
    }
}
