/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;

public class Requirement extends TreeNode implements Comparable<TreeNode> {
    private static final String TYPE_NAME = "Requirement";
    private static final String ATTR_NAME = "_name";
    private static final String ATTR_LOCATIONS = "_locations";
    private static final String TYPE_ROOT_QID = "/Requirements";
    
    public static String getTypeRootQId() {
        return TYPE_ROOT_QID;
    }

    // Constructor for eclipse instancer
    public Requirement() {
        super(null, null);
    }

    public Requirement(TreeDB db, UUID uuid) {
        super(db, uuid);
        setType(TYPE_NAME);
    }

    @SuppressWarnings("unchecked")
    public List<String> getLocationQIds() {
        Object val = getAttributeValue(ATTR_LOCATIONS);
        if(val == null) {
            return Collections.EMPTY_LIST;
        }
        if (val instanceof List) {
            return (List<String>) getAttributeValue(ATTR_LOCATIONS);
        }
        throw new IllegalStateException(ATTR_LOCATIONS+" value must be string list");
    }

    public List<Location> getLocations() {
        List<Location> result = new ArrayList<Location>();
        for(String locQId: getLocationQIds()) {
            TreeNode locNode = getTreeDB().getNode(locQId);
            if( locNode != null && locNode instanceof Location )
            {
                result.add((Location)locNode);
            }
        }
        return Collections.unmodifiableList(result);
    }

    public Location getLocation(String QId) {
        if(getLocationQIds().contains(QId)) {
            return (Location)getTreeDB().getNode(QId);
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    public void addLocation(String nodeQId) {
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null ) {
            locations = new ArrayList<String>();
        }
        locations.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_LOCATIONS,
                locations));
        // destroy locations label cache
        for(String l : locations) {
        	Location loc = (Location) getTreeDB().getNode(l);
        	if(loc != null) { // Can be if node not created still
        		loc.resetLabel();
        	}
        }
    }

    @SuppressWarnings("unchecked")
    public boolean removeLocation(String nodeQId) {
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null || !locations.remove(nodeQId)) {
            return false;
        }
        putAttribute(new Attribute(AttributeType.LIST, ATTR_LOCATIONS,
                locations));
        // destroy locations label cache
        updateLocations();
        return true;
    }

    public void setName(String name) {
        if(getId().endsWith(name)) {
            removeAttribute("_name");
        } else {
            putAttribute(new Attribute(AttributeType.STRING, ATTR_NAME, name));
        }
        updateLocations();
    }

    @Override
    public void setId(String id) {
        if(getQualifiedId().equals(Requirement.getTypeRootQId())) {
            throw new UnsupportedOperationException("Can't rename root requirement");
        }
        updateLocations();
        super.setId(id);
    }

    private void updateLocations() {
        for(String l : getLocationQIds()) {
            Location loc = (Location) getTreeDB().getNode(l);
            if(loc != null) { // Can be if node not created still
                loc.resetLabel();
            }
        }
    }

    public String getName() {
        if (getAttributeValue(ATTR_NAME) == null)
            return getId();
        return (String) getAttributeValue(ATTR_NAME);
    }

    @Override
    public String toString() {
        return getId() + "(" + getUUId().toString() + ")";
    }

    @Override
    public int compareTo(TreeNode o) {
        if (o instanceof Requirement) {
            Requirement req = (Requirement) o;
            return getName().compareTo(req.getName());
        }
        else if(o instanceof TreeNode){
            TreeNode req = (TreeNode) o;
            return getName().compareTo(req.getId());
        }
        throw new IllegalArgumentException();
    }

    public String getUserFriendlyName() {
        if(getId().equals("Requirements")) return "";
        if(getName()!=null && !getId().equals(getName())) return getName();
        Requirement parentReq = (Requirement) getParent();
        if(parentReq.getUserFriendlyName().length()>0) {
            return parentReq.getUserFriendlyName()+"/"+getId();
        }
        return getId();
    }
}
